/* drivers/barcelona/gpio/gpio.c
 *
 * Implementation of the GPIO driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Authors: Dimitry Andric <dimitry.andric@tomtom.com>
 *          Koen Martens <kmartens@sonologic.nl>
 *          Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <barcelona/Barc_Gpio.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>

/* Defines */
#define PFX "gpio: "
#define PK_DBG PK_DBG_FUNC

#define TOMTOM_USB_VBUS_DETECT

#define GPIO_POLL_DELAY (HZ / 5)

extern int btTxTimeout;

/* Forward declarations */
static void gpio_status_poll(unsigned long data);

/* Globals */
static HARDWARE_STATUS gpio_hw_status;
static unsigned int diskAccessStarted = 0;
static unsigned int btModeState;
static struct timer_list gpio_timer = TIMER_INITIALIZER(gpio_status_poll, 0, 0);
static int power_button_timer = -1;
static int ignition_timer = -1;
static int diskAccess = 0;
static atomic_t uart_break_cnt;
static int uart_low_cnt = 0;
static int uart_high_cnt = 0;

static void gpio_tasklet(unsigned long data);
DECLARE_TASKLET(gpio_tasklet_handle, gpio_tasklet, 0L);

/* Functions */
void gpio_handle_break(int uart)
{
		/* only support for UART0 */
		if (uart == 0) 
		{
			atomic_inc(&uart_break_cnt);
		}
		return;
}

static void gpio_tasklet(unsigned long data)
{
	// Store status now so it can be read by the ioctl

	IO_Update(1);

	gpio_hw_status.u8DockStatus = IO_GetDockState();

	if (IO_GetInput(LIGHTS_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_LIGHTS_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_LIGHTS_MASK;
	if (IO_GetInput(IGNITION)) gpio_hw_status.u8InputStatus |= DOCK_IGNITION_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_IGNITION_MASK;
	if (IO_GetInput(EXTMIC_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_EXTMIC_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_EXTMIC_MASK;
	if (IO_GetInput(HEADPHONE_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_HEADPHONE_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_HEADPHONE_MASK;
	if (IO_GetInput(LINEIN_DETECT)) gpio_hw_status.u8InputStatus |= DOCK_LINEIN_MASK; else gpio_hw_status.u8InputStatus &= ~DOCK_LINEIN_MASK;
	if (IO_GetInput(USB_HOST_DETECT)) gpio_hw_status.u8InputStatus |= USB_DETECT_MASK; else gpio_hw_status.u8InputStatus &= ~USB_DETECT_MASK;
	
	/* UART RXD line test for valencia and other extended dock users - NEEDS PULLDOWN (2442) to work */
	/* optionally, for interupt-driven detection, a parallel interrupt input RXD_DOCK_INT to be there */
	if (IO_HaveRxdDetect()) 
	{
		if (gpio_hw_status.u8InputStatus & DOCK_RXD_DETECT_MASK) 
		{
			int break_cnt = 0;
			/* rxd was previously detected (high) */
			/* check if any breaks occurred in UART0 driver */
			while(atomic_sub_and_test(1, &uart_break_cnt)) 
			{
				break_cnt++;
			}
			if (break_cnt) 
			{
				IO_SetInput(RXD_DOCK);
				if (IO_GetInput(RXD_DOCK) == 0) 
				{
					/* uart still low, good */
					gpio_hw_status.u8InputStatus &= ~DOCK_RXD_DETECT_MASK;
				} 
				else 
				{
					/* break detected, but line still high - let it debounce, uart will count another break */
					IO_SetFunction(RXD_DOCK);
				}
			} 
			else 
			{
				/* no break counted, line assumed to be busy, most of the time high */
				if (IO_GetInput(RXD_DOCK) == 0) 
				{
					uart_low_cnt++;
					if (uart_low_cnt > 5) 
					{
						// ca 1 sec low, disconnected
						gpio_hw_status.u8InputStatus &= ~DOCK_RXD_DETECT_MASK;
					}
				} 
				else 
				{
					uart_low_cnt = 0;
				}
			}
		} 
		else 
		{
			/* rxd was low (disconnected), check if came up */
			IO_SetInput(RXD_DOCK);
			if (IO_GetInput(RXD_DOCK) == 0) 
			{
				/* still low, good */
				uart_high_cnt = 0;
			} 
			else 
			{
				/* line assumed to be busy or still high */
				if (uart_high_cnt < 5) 
				{
					uart_high_cnt++;
				} 
				else 
				{
					IO_SetFunction(RXD_DOCK);
					gpio_hw_status.u8InputStatus |= DOCK_RXD_DETECT_MASK;
				}
			}
		}
	} 
	else 
	{
		/* no serial detect dock */
		/* serial always active */
		gpio_hw_status.u8InputStatus |= DOCK_RXD_DETECT_MASK;
	}
	
	// Auto-switch off if in dock and power has been off for more than five seconds
	// Check if docked
	if( ( IO_GetModelId() != GOTYPE_VALENCIA500 ) && ( IO_GetModelId() != GOTYPE_VALENCIA700 ) && IO_GetInput( DOCK_SENSE ) ) 
	{
		// External power?
		if (IO_GetInput(ACPWR)) 
		{
			// Enable running timer when external power is removed (wait 5 seconds)
			ignition_timer = 25;
		}
		else
		{
			// Count down timer when external power was active while docked
			if (ignition_timer > 0)
			{
				ignition_timer--;
			}
			else
			if (ignition_timer == 0)
			{
				// Disable timer to wait for connection of external power
				ignition_timer = -1;
				// Send message to application to shut down
				gpio_hw_status.u8InputStatus |= ONOFF_MASK;
			}
		}
	}
	else 
	{
		// Not docked, so disable timer
		ignition_timer = -1;
	}
	
	// Button pressed?
	if (IO_GetInput(ON_OFF)) 
	{
		// Count down timer when button was released
		if (power_button_timer > 0)
		{
			power_button_timer--;
		}
		else
		if (power_button_timer == 0)
		{
			// Disable timer to wait for button release
			power_button_timer = -1;
			// Send message to application to shut down
			gpio_hw_status.u8InputStatus |= ONOFF_MASK;
		}
	}
	else
	{
		// Enable running timer when button is released (wait 1/5 second)
		power_button_timer = 1;
	}
	
	// HDD access?
	if (IO_GetInput(HDD_LED))
	{
		// Transition to disk access started?
		if (diskAccess == 0)
		{
			diskAccess = 1;
			diskAccessStarted = 1;
		}
	} else diskAccess = 0;
	
	// Power state
	if (IO_GetInput(ACPWR))
	{
		if (IO_GetInput(CHARGEFAULT))
		{
			gpio_hw_status.u8ChargeStatus = CHARGE_STATE_FAULT;
		}
		else if (IO_GetInput(CHARGING))
		{
			IO_Activate(CHARGE_OUT);
			if (IO_GetInput(CHARGING))
			{
				gpio_hw_status.u8ChargeStatus = CHARGE_STATE_TOPPINGUP;
			}
			else
			{
				gpio_hw_status.u8ChargeStatus = CHARGE_STATE_FAST_CHARGING;
			}
			IO_Deactivate(CHARGE_OUT);
		}
		else
		{
			gpio_hw_status.u8ChargeStatus = CHARGE_STATE_COMPLETE;
		}
	}
	else
	{
		gpio_hw_status.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
	}
}

static void gpio_status_poll(unsigned long data)
{
	tasklet_schedule(&gpio_tasklet_handle);
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
}

static int gpio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	
	switch (cmd) {
	case IOR_HWSTATUS:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_to_user((void __user *) arg, &gpio_hw_status, sizeof gpio_hw_status) ? -EFAULT : 0;
		tasklet_enable(&gpio_tasklet_handle);
		break;

	case IOR_DISK_ACCESS:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_to_user((void __user *) arg, &diskAccessStarted, sizeof diskAccessStarted) ? -EFAULT: 0;
		diskAccessStarted = 0;
		tasklet_enable(&gpio_tasklet_handle);
		break;
	
	case IOR_BT_MODE:
		tasklet_disable(&gpio_tasklet_handle);
		ret = copy_from_user(&btModeState, (void __user *) arg, sizeof btModeState) ? -EFAULT: 0;
		tasklet_enable(&gpio_tasklet_handle);
		/* set BT mode */
		if (btModeState & 0x01)
			IO_Activate(BT_MODE);
		else
			IO_Deactivate(BT_MODE);
		/* issue BT reset */
		IO_Activate(BT_RESET);
		mdelay(8); /* wait at least 5ms according to BC4ROM datasheet */
		IO_Deactivate(BT_RESET);
		mdelay(8);
		IO_Activate(BT_RESET);
		mdelay(8);
		IO_Deactivate(BT_RESET);
		mdelay(100);
		break;

	case IOR_GET_BT_ERROR:
		ret = copy_to_user((void __user *) arg, &btTxTimeout, sizeof btTxTimeout) ? -EFAULT : 0;
		break;

	case IOR_SET_BT_ERROR:
		ret = copy_from_user(&btTxTimeout, (void __user *) arg, sizeof btTxTimeout) ? -EFAULT: 0;
		break;

	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int gpio_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations gpio_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= gpio_ioctl,
	.open		= gpio_open,
	.release	= gpio_release,
};

static void gpio_hw_init(void)
{
	IO_SetInput(ON_OFF);
	IO_SetInput(IGNITION);
	IO_SetInput(ACPWR);
	IO_SetInput(CHARGEFAULT);
	IO_SetInput(CHARGING);
	IO_SetInput(LIGHTS_DETECT);
	IO_SetInput(HEADPHONE_DETECT);
	IO_SetInput(EXTMIC_DETECT);
	IO_SetInput(LINEIN_DETECT);
	IO_SetInput(DOCK_SENSE);
	IO_SetInput(DOCK_DESK_SENSE);
	IO_SetInput(USB_HOST_DETECT);
	IO_SetInput(RXD_DOCK_INT);

	// Stop timers
	power_button_timer = -1;
	ignition_timer = -1;
	// Clear state from before suspend
	gpio_hw_status.u8InputStatus = 0;
	atomic_set(&uart_break_cnt, 0);
	
	// Get new state
	tasklet_schedule(&gpio_tasklet_handle);
	mod_timer(&gpio_timer, jiffies + GPIO_POLL_DELAY);
}

static void gpio_hw_exit(void)
{
	del_timer_sync(&gpio_timer);
}

static int gpio_probe(struct device *dev)
{
	int ret;

	gpio_hw_init();

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(GPIO_MAJOR, GPIO_DEVNAME, &gpio_fops);
	if (ret < 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", GPIO_MAJOR, ret);
		return ret;
	}

	PK_DBG("Done\n");
	return 0;
}

static int gpio_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(GPIO_MAJOR, GPIO_DEVNAME);

	gpio_hw_exit();

	PK_DBG("Done\n");
	return 0;
}

static void gpio_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	gpio_hw_exit();
}

#ifdef CONFIG_PM

static int gpio_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		gpio_hw_exit();
	}
	return 0;
}

static int gpio_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		gpio_hw_init();
	}
	return 0;
}

#else /* CONFIG_PM */
#define gpio_suspend NULL
#define gpio_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver gpio_driver = {
	.name		= "tomtomgo-gpio",
	.bus		= &platform_bus_type,
	.probe		= gpio_probe,
	.remove		= gpio_remove,
	.shutdown	= gpio_shutdown,
	.suspend	= gpio_suspend,
	.resume		= gpio_resume,
};

static int __init gpio_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO GPIO Driver, (C) 2004,2005 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&gpio_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit gpio_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&gpio_driver);
	PK_DBG("Done\n");
}

module_init(gpio_mod_init);
module_exit(gpio_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com> and Koen Martens <kmartens@sonologic.nl>");
MODULE_DESCRIPTION("TomTom GO GPIO Driver");
MODULE_LICENSE("GPL");

/* EOF */
