#include <linux/kernel.h>
#include <linux/err.h>

#include <asm/errno.h>
#include <asm/io.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-iis.h>
#include <asm/hardware/clock.h>

#include <barcelona/Barc_snd.h>
#include <barcelona/gopins.h>
#include <barcelona/debug.h>

#include "iis.h"
#include "codec.h"

#define PFX "iis: "
#define PK_DBG PK_DBG_FUNC

#define rMISCCR     (*(volatile unsigned *)S3C2410_MISCCR)
#define rDCLKCON    (*(volatile unsigned *)S3C2410_DCLKCON)

struct iis_prescaler_info {
	const unsigned long hz;
	const unsigned long cdclk_goal;
	unsigned div;
	unsigned long res;
};

struct iis_regs {
	unsigned long iismod;
	unsigned long iisfcon;
	unsigned long iispsr;
	unsigned long iiscon;
};

static struct clk *iis_clock;
static unsigned iis_clockrate;
static struct iis_regs iis_regs;
static unsigned current_samplerate;
static unsigned current_samplesize;

void iis_showbus(const char *message)
{
	unsigned long mod,fcon,psr,con;

	mod  = readl(S3C2410_IISMOD);
	fcon = readl(S3C2410_IISFCON);
	psr  = readl(S3C2410_IISPSR);
	con  = readl(S3C2410_IISCON);


	printk("(%s) iismod: %lx, iisfcon: %lx, iiscon: %lx, iispsr: %lx\n",
	       message, mod, fcon, con, psr);
}

void iis_suspend(void)
{
	iis_regs.iismod  = readl(S3C2410_IISMOD);
	iis_regs.iisfcon = readl(S3C2410_IISFCON);
	iis_regs.iispsr  = readl(S3C2410_IISPSR);
	iis_regs.iiscon  = readl(S3C2410_IISCON);
	codec_suspend();
}

void iis_resume(void)
{
	codec_resume();
	writel(iis_regs.iismod,  S3C2410_IISMOD);
	writel(iis_regs.iisfcon, S3C2410_IISFCON);
	writel(iis_regs.iispsr,  S3C2410_IISPSR);
	writel(iis_regs.iiscon,  S3C2410_IISCON);
}

void iis_control_cmd(enum iis_cmd command)
{
	unsigned iiscon, iisfcon;

	switch (command) {
	case IIS_IDLE_TX:
		IO_Deactivate(I2SSDO);
		break;

	case IIS_ACTIVE_TX:
		IO_SetFunction(I2SSDO);
		break;

	case IIS_ALIGN_TX:
		if(readl(S3C2410_IISCON) & S3C2410_IISCON_LRINDEX) {
			iisfcon = readl(S3C2410_IISFCON);
			writel(iisfcon & (~S3C2410_IISFCON_TXDMA), S3C2410_IISFCON);
			writew(0, S3C2410_IISFIFO);
			writel(iisfcon, S3C2410_IISFCON);
		}
		break;


	case IIS_FLUSH_TX:
		iisfcon = readl(S3C2410_IISFCON);
		iisfcon &= ~S3C2410_IISFCON_TXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		iisfcon |=  S3C2410_IISFCON_TXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		break;

	case IIS_FLUSH_RX:
		iisfcon = readl(S3C2410_IISFCON);
		iisfcon &= ~S3C2410_IISFCON_RXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		iisfcon |=  S3C2410_IISFCON_RXENABLE;
		writel(iisfcon, S3C2410_IISFCON);
		break;

	case IIS_START:
		iiscon = readl(S3C2410_IISCON);
		iiscon |= S3C2410_IISCON_IISEN;
		writel(iiscon, S3C2410_IISCON);
		break;

	case IIS_STOP:
		iiscon = readl(S3C2410_IISCON);
		iiscon &= ~S3C2410_IISCON_IISEN;
		writel(iiscon, S3C2410_IISCON);
		break;

	default:
		break;

	}
}

static inline unsigned divrnd(unsigned nom, unsigned den)
{
	return (nom + den / 2) / den;
}

int iis_set_samplerate(unsigned rate)
{
	/* Formula is: div + 1 = iisclk / (fs * rate) */
	unsigned div1;
	unsigned div2;
	unsigned realrate1;
	unsigned realrate2;
	unsigned diff1;
	unsigned diff2;
	enum codec_fs fs;
	unsigned div;
	unsigned realrate;
	unsigned long flags;
	unsigned long iismod;

	if (IO_GetCodecMaster()) 
	{
		// Output DCLK1 on CLKOUT1
		rMISCCR = (rMISCCR & (~(7 << 8))) | (5 << 8);
		// DCLK1 = UCLK / 4
		rDCLKCON = (1 << 24) | (3 << 20) | (1 << 17) | (1 << 16);

		IO_SetInput(CDCLK);
		IO_SetFunction(CDCLK_12MHZ);

		div = divrnd(iis_clockrate, 12000000) - 1;
		current_samplerate = rate;
		codec_set_fs(0,rate);
	}
	else
	{
		IO_SetInput(CDCLK_12MHZ);
		IO_SetFunction(CDCLK);

		div1 = divrnd(iis_clockrate, FS256 * rate);
		div2 = divrnd(iis_clockrate, FS384 * rate);
		if (unlikely((div1 == 0 || div1 > 32) || (div2 == 0 || div2 > 32))) {
			PK_ERR("Input rate %u too %s.\n", rate, (div1 == 0 || div2 == 0) ? "low" : "high");
			return -EINVAL;
		}
	
		realrate1 = divrnd(iis_clockrate, FS256 * div1);
		realrate2 = divrnd(iis_clockrate, FS384 * div2);
		if (unlikely((realrate1 < 4000 || realrate1 > 100000) || (realrate2 < 4000 || realrate2 > 100000))) {
			PK_ERR("Weird output rates %u, %u for input rate %u.\n", realrate1, realrate2, rate);
			return -EINVAL;
		}
	
		/* Rate should be okay, store it */
		current_samplerate = rate;
	
		/* Calculate best rate */
		diff1 = (rate <= realrate1) ? realrate1 - rate : rate - realrate1;
		diff2 = (rate <= realrate2) ? realrate2 - rate : rate - realrate2;
	
		PK_DBG("wantrate %u, div1 %u, realrate1 %u, diff1 %u\n", rate, div1, realrate1, diff1);
		PK_DBG("wantrate %u, div2 %u, realrate2 %u, diff2 %u\n", rate, div2, realrate2, diff2);
	
		if (diff1 <= diff2) {
			fs = FS256;
			div = div1;
			realrate = realrate1;
		} else {
			fs = FS384;
			div = div2;
			realrate = realrate2;
		}
	
		div--; /* This is the /real/ divider */
		PK_DBG("Using divider %u, fs %u, wantrate %u, realrate %u\n", div, fs, rate, realrate);

		local_irq_save(flags);
		codec_set_fs(fs,0);
		iismod = readl(S3C2410_IISMOD);
		if (fs == FS256)
			iismod &= ~S3C2410_IISMOD_384FS;
		else
			iismod |= S3C2410_IISMOD_384FS;
		writel(iismod, S3C2410_IISMOD);
		writel((div << 5) | div, S3C2410_IISPSR);
		local_irq_restore(flags);
	}

	//iis_showbus("after set samplerate");

	return 0;
}

unsigned iis_get_samplerate(void)
{
	return current_samplerate;
}

int iis_set_samplesize(unsigned size)
{
	unsigned long iismod = readl(S3C2410_IISMOD);

	if (size == 8) iismod &= ~S3C2410_IISMOD_16BIT;
	else if (size == 16) iismod |= S3C2410_IISMOD_16BIT;
	else return -EINVAL;

	current_samplesize = size;
	writel(iismod, S3C2410_IISMOD);
	//iis_showbus("after setsmp");
	return 0;
}

unsigned iis_get_samplesize(void)
{
	return current_samplesize;
}

int iis_init(void)
{
	unsigned long iismod, iisfcon, iiscon;

	IO_SetInput(CDCLK_12MHZ);
	IO_SetFunction(CDCLK);
	IO_Deactivate(I2SSDO);
	IO_SetFunction(I2SSCLK);
	IO_SetFunction(I2SLRCK);
	IO_SetFunction(I2SSDI);

	iis_clock = clk_get(NULL, "iis");
	if (IS_ERR(iis_clock)) {
		iis_clock = NULL;
		PK_ERR("Failed to get IIS clock\n");
		return -ENOENT;
	}

	codec_init();

	clk_use(iis_clock);
	clk_enable(iis_clock);

	iis_clockrate = clk_get_rate(iis_clock);
	PK_INFO("Current clock rate: %u\n", iis_clockrate);

	iiscon = S3C2410_IISCON_PSCEN |
	         S3C2410_IISCON_IISEN |
	         S3C2410_IISCON_TXDMAEN |
	         S3C2410_IISCON_RXDMAEN;

	iisfcon = S3C2410_IISFCON_RXDMA |
	          S3C2410_IISFCON_TXDMA |
	          S3C2410_IISFCON_TXENABLE |
	          S3C2410_IISFCON_RXENABLE;

	iismod = S3C2410_IISMOD_LR_RLOW |
	         S3C2410_IISMOD_IIS |
	         S3C2410_IISMOD_256FS |
	         S3C2410_IISMOD_32FS |
	         S3C2410_IISMOD_TXRXMODE;
	         
	if (IO_GetCodecMaster()) iismod |= (1 << 8);	         

//	dump_iis_bus("Test1");
	writel(iismod,  S3C2410_IISMOD);
	writel(iisfcon, S3C2410_IISFCON);
	writel(iiscon,  S3C2410_IISCON);
//	dump_iis_bus("Test2");

	iis_set_samplerate(INITIAL_SAMPLERATE);
	iis_set_samplesize(16);

	return 0;
}

void iis_exit(void)
{
	if (iis_clock != NULL) {
		clk_disable(iis_clock);
		clk_unuse(iis_clock);
		clk_put(iis_clock);
		iis_clock = NULL;
	}
}
