#include <asm/arch/regs-gpio.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>

#define PFX "sd_card: "
#define PK_DBG PK_DBG_NONE

#define USE_LBA_IF_SUPPORTED

extern void SDCARDEXT_Init(unsigned reqsdibase, int reqs3c2440, void (*reqmsdelay)(unsigned ms));
extern void SDCARDEXT_BeforePowerOn(void);
extern int SDCARDEXT_AfterPowerOn(void);
extern int SDCARDEXT_WriteSectors(unsigned char *buffer,unsigned int sector,unsigned int number);
extern int SDCARDEXT_ReadSectors(unsigned char *buffer,unsigned int sector,unsigned int block);
extern unsigned int SDCARDEXT_GetAmountSectors(void);
extern unsigned int SDCARDEXT_GetSerial(unsigned char *buffer);

#if 0

#define DUMP_HEX_WIDTH 16

// Dump a hex table
void SDCARD_DumpHex(unsigned char *buffer,int size)
{
	int x,y;

	printk(KERN_INFO);
	for(y=0;y<((size+DUMP_HEX_WIDTH-1)/DUMP_HEX_WIDTH);y++)
	{
		printk("%08X ",y*DUMP_HEX_WIDTH);

		for(x=0;x<DUMP_HEX_WIDTH;x++)
		{
			printk("%02X ",buffer[(x)+y*DUMP_HEX_WIDTH]);
		}
		for(x=0;x<DUMP_HEX_WIDTH;x++)
		{
			if ((buffer[(x)+y*DUMP_HEX_WIDTH] > 31) && (buffer[(x)+y*DUMP_HEX_WIDTH] != 127))
			{
				printk("%c",buffer[(x)+y*DUMP_HEX_WIDTH]);
			}
			else
			{
				printk(".");
			}
		}
		printk("\n");
	}
}

#endif

static void SDCARD_Delay(unsigned ms)
{
	mdelay(ms);
}

static unsigned int sdcardOn = 0;

void SDCARD_Off(void)
{
	if (sdcardOn)
	{
		// Disable sdcard signals
		IO_Deactivate(SDCLK);
		IO_Deactivate(SDCMD);
		IO_Deactivate(SDDATA0);
		IO_Deactivate(SDDATA1);
		IO_Deactivate(SDDATA2);
		IO_Deactivate(SDDATA3);
		IO_Deactivate(SD_PWR_ON);

		sdcardOn = 0;
	}
}

static void SDCARD_On(void)
{
	SDCARD_Off();

	SDCARD_Delay(100);

	// Enable sdcard signals
	IO_Deactivate(SDCLK);
	IO_SetFunction(SDCMD);
	IO_SetFunction(SDDATA0);
	IO_SetFunction(SDDATA1);
	IO_SetFunction(SDDATA2);
	IO_SetFunction(SDDATA3);
	IO_Activate(SD_PWR_ON);
	SDCARD_Delay(5);
	IO_SetFunction(SDCLK);

	SDCARD_Delay(100);

	sdcardOn = 1;
}

unsigned int SDCARD_InSlot(void)
{
#ifdef CONFIG_SMDK2440_BOARD
	// on the evaluation board, checking that SD is present does not work
	return 1;
#else /* CONFIG_SMDK2440_BOARD */
	if (!IO_GetInput(CD_SD))
	{
		PK_WARN("failed!\n");
		SDCARD_Off();
		return 0;
	}
	else
	{
		PK_DBG("success!\n");
		return 1;
	}
#endif /* CONFIG_SMDK2440_BOARD */
}

void SDCARD_Init(void)
{
	int s3c2440 = (IO_GetCpuType() != GOCPU_S3C2410) ? 1 : 0;
	SDCARDEXT_Init((unsigned)S3C24XX_VA_SDI, s3c2440, &SDCARD_Delay);
}

int SDCARD_InitDisk(void)
{
	if (!SDCARD_InSlot())
	{
		PK_WARN("card not in slot!\n");
		return 0;
	}

	SDCARDEXT_BeforePowerOn();

	// Config SD switch input
	IO_SetInterruptOnToggle(CD_SD);
	// Disable write protect input
	IO_Deactivate(WP_SD);

	SDCARD_On();

	if (SDCARDEXT_AfterPowerOn())
	{
		PK_DBG("success!\n");
		return 1;
	}

	PK_WARN("failed!\n");
	SDCARD_Off();

	return 0;
}

int SDCARD_WriteSectors(unsigned char *buffer,unsigned int sector,unsigned int number)
{
	PK_DBG("%u %u\n",sector,number);

	if (!SDCARD_InSlot()) return 0;

	if (SDCARDEXT_WriteSectors(buffer,sector,number)) return 1;

	SDCARD_Off();

	return 0;
}

int SDCARD_ReadSectors(unsigned char *buffer,unsigned int sector,unsigned int number)
{
	PK_DBG("%u %u\n",sector,number);

	if (!SDCARD_InSlot()) return 0;

	if (SDCARDEXT_ReadSectors(buffer,sector,number))
	{
		PK_DBG("success!\n");
		return 1;
	}

	PK_WARN("failed!\n");
	SDCARD_Off();

	return 0;
}

unsigned int SDCARD_GetAmountSectors(void)
{
	if (!SDCARD_InSlot()) return 0;

	return SDCARDEXT_GetAmountSectors();
}

unsigned int SDCARD_GetSerial(unsigned char *buffer)
{
	return SDCARDEXT_GetSerial(buffer);
}

#define IRQ_HOTSWAP IRQ_EINT2

static void (*sdcard_hotswap_func)(int);
static int sdcard_hotswap_device = -1;
static int sdcard_hotswap_irq;

irqreturn_t SDCARD_HotswapIrq(int irqno, void *dev, struct pt_regs *regs)
{
	if (unlikely(sdcard_hotswap_func == NULL))
	{
		PK_WARN("IRQ with NULL hotswap func\n");
	}
	else if (unlikely(sdcard_hotswap_device == -1))
	{
		PK_WARN("IRQ with invalid hotswap dev\n");
	}
	else
	{
		sdcard_hotswap_func(sdcard_hotswap_device);
	}

	return IRQ_HANDLED;
}

int SDCARD_HotswapInit(void (*func)(int), int device)
{
	int ret;

	// store function & devno
	sdcard_hotswap_func = func;
	sdcard_hotswap_device = device;

	// request irq
	ret = request_irq(IO_GetInterruptNumber(CD_SD), SDCARD_HotswapIrq, 0, "sdcard hotswap" , NULL);
	if (ret != 0)
	{
		sdcard_hotswap_func = NULL;
		sdcard_hotswap_device = -1;
		PK_ERR("Failed to request SD hotswap IRQ (%d)\n", ret);
		return 1;
	}

	sdcard_hotswap_irq = 1;
	return 0;
}

void SDCARD_HotswapUninit(void)
{
	if (sdcard_hotswap_irq != 0)
	{
		free_irq(IRQ_HOTSWAP, NULL);
		sdcard_hotswap_irq = 0;
	}
}
